#include <windows.h>
#include <gl/gl.h>

#include "cmodel.h"

void CModel::Reset(void)
{
	vtxCount = 0;
	vtx = NULL;

	planeCount = 0;
	edgeCount = 0;

	plane = NULL;
	edge = NULL;
}

void CModel::Unload(void)
{
	if (vtx) free(vtx);
	if (plane) free(plane);
	if (edge) free(edge);

	Reset();
}


#include <stdio.h>
#include "error.h"
#include "const.h"
bool CModel::Load(char *filename)
{
	int file_vtxCount;
	int file_triCount;
	int t,v;

	int *tri;

	//CUSTOM MDL FILE FORMAT

	strcpy(title, filename);

	char sig[4];

	FILE *File = fopen(filename, "rb");
	if (File == NULL)
	{
		Error("%s: unable to open model file", filename);
		return false;
	}

	fread(sig, 1, 4, File);
	if (memcmp(sig, "MMDL", 4))
	{
		Error("%s: bad signature on file", filename);
		return false;
	}
	fread(&file_vtxCount, sizeof(int), 1, File);
	fread(&file_triCount, sizeof(int), 1, File);

	if (file_vtxCount < 0)
	{
		Error("%s: cmodel has bad vertex count: %d", filename, file_vtxCount);
		fclose(File);
		return false;
	}
	if (file_triCount < 0)
	{
		Error("%s: cmodel has bad plane count: %d", filename, file_triCount);
		fclose(File);
		return false;
	}

	//allocate verticies
	if ((vtx = (Vector *)malloc(sizeof(Vector) * file_vtxCount))==NULL)
	{
		Error("%s: unable to allocate vertex array", filename);
		fclose(File);
		return false;
	}
	//read verticies
	if (fread(vtx, sizeof(Vector), file_vtxCount, File) != (size_t)(file_vtxCount))
	{
		Error("%s: unable to read vertex array", filename);
		free(vtx);
		fclose(File);
		return false;
	}
	//skip texture coordinates
	if (fseek(File, sizeof(float)*file_vtxCount*2, SEEK_CUR))
	{
		Error("%s: unable to skip texture coordinate array", filename);
		free(vtx);
		fclose(File);
		return false;
	}
	//allocate triangles
	if ((tri = (int *)malloc(sizeof(int) * 3 * file_triCount))==NULL)
	{
		Error("%s: unable to allocate tri array", filename);
		free(vtx);
		fclose(File);
		return false;
	}
	//read triangles
	if (fread(tri, sizeof(int)*3, file_triCount, File) != (size_t)(file_triCount))
	{
		Error("%s: unable to read tri array", filename);
		free(vtx);
		free(tri);
		fclose(File);
		return false;
	}

	fclose(File);

	if ((plane = (Plane *)malloc(sizeof(Plane) * file_triCount))==NULL)
	{
		Error("%s: unable to allocate plane array", filename);
		free(tri);
		free(vtx);
		fclose(File);
		return false;
	}
	for (v = 0; v < file_triCount; v++)
	{
		t = v*3;
		plane[v].pos = vtx[tri[t]];
		plane[v].normal.CalcNormal(&vtx[tri[t]], &vtx[tri[t+1]], &vtx[tri[t+2]]);
	}

	//calc bbox
	bbox.min = Vector(fInf,fInf,fInf);
	bbox.max = Vector(-fInf,-fInf,-fInf);
	radius = -1;
	for (v = 0; v < file_vtxCount; v++)
	{
		//stretch bbox
		bbox.Stretch(&vtx[v]);

		//stretch radius
		float l = vtx[v].length();
		if (l > radius) radius = l;
	}

	//calc edgeCount
	edgeCount = 0;
	for (int a = 0; a < file_triCount-1; a++)
	{
		for (int b = a+1; b < file_triCount; b++)
		{
			int count = 0;

			//if a and b share two verticies - create an edge
			for (v = 0; v < 3; v++)	//verticies of tri a
			{
				for (t = 0; t < 3; t++)	//verticies of tri b
				{
					if (tri[3*a+v] == tri[3*b+t]) count++;
				}
			}

			if (count == 2) edgeCount++;
		}
	}
	//allocate edge buffer
	if ((edge = (int *)malloc(2*edgeCount * sizeof(int)))==NULL)
	{
		Error("%s: unable to allocate edge array", filename);
		free(vtx);
		free(tri);
		free(plane);
		return false;
	}
	//create edges
	int l = 0;
	for (int a = 0; a < file_triCount-1; a++)
	{
		for (int b = a+1; b < file_triCount; b++)
		{
			int e[2];
			int count = 0;

			//if a and b share two verticies - create an edge
			for (v = 0; v < 3; v++)	//verticies of tri a
			{
				for (t = 0; t < 3; t++)	//verticies of tri b
				{
					if (tri[3*a+v] == tri[3*b+t])
					{
						if (count < 2) e[count] = tri[3*a+v];
						count++;
					}
				}
			}

			if (count == 2)
			{
				edge[l] = e[0]; l++;
				edge[l] = e[1]; l++;
			}
		}
	}
	free(tri);
	vtxCount = file_vtxCount;
	planeCount = file_triCount;

	return true;
}

CModel *GetCModel(Chain *p, char *title)
{
	CModel *m = (CModel *)p->ChainHead();
	for (; m; m = (CModel *)m->ChainNext())
		if (!_stricmp(title, m->title)) break;

	return m;
}
