#include <windows.h>
#include <gl/gl.h>

#include "camera.h"
#include "display.h"
#include "console.h"

#include "main.h"	//for main_display

// ********** CONSOLE **********

M_CVar *cvar_cam[8];

void CamInitCVars(void)
{
	static bool init_d = false;
	if (!init_d)
	{
		init_d = true;

			//the console variables - create the camera console vars before you set the camera
		cvar_cam[0] =	new M_CVar((void *)NULL,"CamPos");
		cvar_cam[1] =	new M_CVar((void *)NULL,"CamAngle");
		cvar_cam[2] = 	new M_CVar((void *)NULL,"CamTrackMethod");
		cvar_cam[3] = 	new M_CVar((void *)NULL,"CamTrackSlerpC");
		cvar_cam[7] =	new M_CVar((void *)NULL,"CamTrackLerpC");
		cvar_cam[4] = 	new M_CVar((void *)NULL,"CamChaseOffset");
		cvar_cam[5] = 	new M_CVar((void *)NULL,"CamZNear");
		cvar_cam[6] = 	new M_CVar((void *)NULL,"CamZFar");
	}
}

// ********** CONSTRUCTION ********** 

M_Camera::M_Camera()
{
	//link all CVars - dont do this outside of any method
	//instead, let the parent chain of the cvar be created and constructed
	//before linking anything to it
	CamInitCVars();

	//upon construction of the position vector and angle quaternion, identity
	//( [0,0,0] and {[0,0,0],1}, respectively) is set

	trackPosition = NULL;
	trackAngle = NULL;
	trackMethod = CAM_TRACK_NO;
	trackSlerpC = 0.5f;	//the slerp coefficient from the current to the tracking angle
	trackLerpC = 0.5f;	//lerp position coeff
	chaseOffset = Vector(-25, 0, 5);

	scale.x = scale.y = scale.z = 1.f;
	clip_plane_active = false;

	znear = 1.0f;
	zfar = 30000.0f;

	active = false;
}

M_Camera::M_Camera(M_Camera *src)		//create a cam from another cam
{
	//i hope it wont mind the private variables
	memcpy(this, src, sizeof(M_Camera));
}


bool M_Camera::Init(void) {return true;}

void M_Camera::Activate(void)
{
	//rematch the console variables
	if (cvar_cam[0]) cvar_cam[0]->Reset(&position);
	if (cvar_cam[1]) cvar_cam[1]->Reset(&angle);
	if (cvar_cam[2]) cvar_cam[2]->Reset(&trackMethod);
	if (cvar_cam[3]) cvar_cam[3]->Reset(&trackSlerpC);
	if (cvar_cam[7]) cvar_cam[7]->Reset(&trackLerpC);
	if (cvar_cam[4]) cvar_cam[4]->Reset(&chaseOffset);
	if (cvar_cam[5]) cvar_cam[5]->Reset(&znear);
	if (cvar_cam[6]) cvar_cam[6]->Reset(&zfar);

	active = true;
}

// ********** DESTRUCTION ********** 

void M_Camera::Kill(void) {}

// ********** SCENE TRANSFORMATION ********** 

void M_Camera::TransformScene(int flags)
{
	//only call this when the viewport dimensions have changed
	if (flags & CAMERA_RESET_PROJECTION)
		SetProjection();	//sets the field of vision / clip planes / etc

	if (flags & (CAMERA_RESET_TRANSLATION | CAMERA_RESET_ROTATION))
		SetModelview(flags);		//sets the translation / rotation
}

// **************************************
// Projection Matrix transformation calls
// **************************************

void M_Camera::SetProjection(void)
{
	float ratio = 0;

	//include a main object *here*
	if (main_display) ratio = (float)main_display->GetHeight() / (float)main_display->GetWidth();

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	//grab that frustum
	glFrustum(-znear, znear, -znear*ratio, znear*ratio, znear, zfar);
}

// *************************************
// Modelview Matrix transformation calls
// *************************************

void M_Camera::SetModelview(int flags)
{
	glMatrixMode(GL_MODELVIEW);

	static float RotMatrix[] = {	0.0,	0.0,	-1.0,	0.0,
									-1.0,	0.0,	0.0,	0.0,
									0.0,	1.0,	0.0,	0.0,
									0.0,	0.0,	0.0,	1.0};

	//set initial conversion
	glLoadMatrixf(RotMatrix);
	//x faces forward
	//y faces to the left
	//z faces upwards

	if (flags & CAMERA_RESET_ROTATION)		RotateScene();
	if (flags & CAMERA_RESET_TRANSLATION)	TranslateScene();
	if (flags & CAMERA_RESET_SCALE)			ScaleScene();

	if ((clip_plane_active) && (flags & (CAMERA_RESET_ROTATION | CAMERA_RESET_TRANSLATION | CAMERA_RESET_SCALE)))
	{
		double clip_plane_eq[4];
		if (flags & CAMERA_RESET_TRANSLATION)
		{
			clip_plane.CalculateEquation(clip_plane_eq);
			glClipPlane(GL_CLIP_PLANE0, clip_plane_eq);
		}
		else	//zero the clip plane just in case?
		{
			clip_plane_eq[0] = clip_plane_eq[1] = clip_plane_eq[2] = clip_plane_eq[3] = 0;
			glClipPlane(GL_CLIP_PLANE0, clip_plane_eq);
		}
		glEnable(GL_CLIP_PLANE0);
	}
}

void M_Camera::TranslateScene(void)
{
	//and translate it
	glTranslatef(-position.x, -position.y, -position.z);
}

void M_Camera::RotateScene(void)
{
	Quat q;
	//then rotate the camera
	q.QuatToAngleAxis(&angle);
	glRotatef(q.w, -q.x, -q.y, -q.z);
}

void M_Camera::ScaleScene(void)
{
	glScalef(scale.x, scale.y, scale.z);
}

// *************************
// Camera clip plane methods
// *************************

void M_Camera::EnableClipPlane(Plane *plane)
{
	clip_plane = *plane;
	clip_plane_active = true;
}

void M_Camera::DisableClipPlane(void)
{
	clip_plane_active = false;
	
	/*
	if the scene had been rendered after enabling the camera clip plane
	this this would have been enabled - so disable it
	*/
	glDisable(GL_CLIP_PLANE0);
}

// **************************
// Camera frustum calculation
// **************************

//calculates the frustum verticies
//the three axii (x/y/z)
//the six border planes
void M_Camera::CalculateFrustum(void)
{
	float ratio = 0;
	//include a main object *here*
	if (main_display) ratio = (float)main_display->GetHeight() / (float)main_display->GetWidth();

//and recalculate the frustum:
//this is here right now because we have already referenced 'ratio'
	/*
	  LR(1) UD(2)  fwd/back(0)
	0	-	-	-
	1	+	-	-
	2	+	+	-
	3	-	+	-
	0	-	-	+
	1	+	-	+
	2	+	+	+
	3	-	+	+
	*/

	/*
	frustum_vtx[0] = position + axis[0] * znear - axis[1] - ratio * axis[2];
	frustum_vtx[1] = position + axis[0] * znear + axis[1] - ratio * axis[2];
	frustum_vtx[2] = position + axis[0] * znear + axis[1] + ratio * axis[2];
	frustum_vtx[3] = position + axis[0] * znear - axis[1] + ratio * axis[2];
	frustum_vtx[4] = position + axis[0] * zfar - axis[1] - ratio * axis[2];
	frustum_vtx[5] = position + axis[0] * zfar + axis[1] - ratio * axis[2];
	frustum_vtx[6] = position + axis[0] * zfar + axis[1] + ratio * axis[2];
	frustum_vtx[7] = position + axis[0] * zfar - axis[1] + ratio * axis[2];
	*/

	static int axis_flag[8][3] = {
		{0, 0, 0},
		{0, 0, 1},
		{0, 1, 1},
		{0, 1, 0},
		{1, 0, 0},
		{1, 0, 1},
		{1, 1, 1},
		{1, 1, 0} };

	//change the axis array to a rotation matrix ?
	axis[0] = Vector(1,0,0);
	axis[1] = Vector(0,1,0);
	axis[2] = Vector(0,0,1);

	for (int v = 0; v < 3; v++) angle.Rotate(&axis[v], &axis[v]);

	for (int v = 0; v < 8; v++)
	{
		Vector vec[3];

		vec[0] = axis[0];
		vec[1] = axis[1];
		vec[2] = axis[2];

		if (!axis_flag[v][1])	vec[1].scale(-1);

		if (!axis_flag[v][2])	vec[2].scale(-ratio);
		else					vec[2].scale(ratio);

		if (!axis_flag[v][0])
		{
			vec[0].scale(znear);
			vec[1].scale(znear);
			vec[2].scale(znear);
		}
		else
		{
			vec[0].scale(zfar);
			vec[1].scale(zfar);
			vec[2].scale(zfar);
		}

		frustum_vtx[v] = position;
		frustum_vtx[v].add(&vec[0]);
		frustum_vtx[v].add(&vec[1]);
		frustum_vtx[v].add(&vec[2]);

		if (v == 0)
		{
			xmin = xmax = frustum_vtx[v].x;
			ymin = ymax = frustum_vtx[v].y;
		}
		else
		{
			if (frustum_vtx[v].x < xmin) xmin = frustum_vtx[v].x;
			if (frustum_vtx[v].x > xmax) xmax = frustum_vtx[v].x;
			if (frustum_vtx[v].y < ymin) ymin = frustum_vtx[v].y;
			if (frustum_vtx[v].y > ymax) ymax = frustum_vtx[v].y;
		}
	}

	//calculate the frustum planes

	/*
	0	znear
	1	zfar
	2	bottom
	3	left
	4	top
	5	right
	*/

	//quick front and back - normal is already same as camera dir
	frustum_plane[0].pos = frustum_vtx[0];
	frustum_plane[0].normal = Vector(-axis[0].x, -axis[0].y, -axis[0].z);

	frustum_plane[1].pos = frustum_vtx[4];
	frustum_plane[1].normal = axis[0];

	for (int v = 0; v < 4; v++)
	{
		int v_inc = (v+1)&3;
		frustum_plane[v+2].pos = Vector(0,0,0);

		frustum_plane[v+2].pos.add(&frustum_vtx[v]);
		frustum_plane[v+2].pos.add(&frustum_vtx[v_inc]);
		frustum_plane[v+2].pos.add(&frustum_vtx[v_inc+4]);
		frustum_plane[v+2].pos.add(&frustum_vtx[v+4]);
		frustum_plane[v+2].pos.scale(0.25f);

		frustum_plane[v+2].normal.CalcNormal(&frustum_vtx[v], &frustum_vtx[v_inc], &frustum_vtx[v_inc+4]);
	}
}

/*
Note about AddRotation():

normal quaternion rotation combination is as follows
q' = q2 * q
this adds the rotation of q2 to q, and would be represent as the following code:
angle.Mul(q, &angle);

the camera quaternion is opposite of that (transposed) because when camera rotations
are added to the current camera angle, the rotations are done relative to the current
orientation of the camera instead of the identity orientation, as is implied in the
previous example.  because of this special case for camera rotations, the products
are multiplied backwards (resulting in a negative rotation?)
*/


// ********** TRACKING ********** 

void M_Camera::Update(void)
{
	//recalculate the position and angle

	switch (trackMethod)
	{
		//ignore CAM_TRACK_NO
	case (CAM_TRACK_YES):
		if (trackAngle) angle = *trackAngle;

		if (trackPosition) position = *trackPosition;
		break;

	case (CAM_TRACK_CHASE):
		//SLERP from the current angle to the track angle
		if (trackAngle) angle.Slerp(&angle, trackAngle, trackSlerpC);

		if (trackPosition)
		{
			//rotate chase offset according to the current angle
			//add it to the tracking position
			//and you got your current position
			Vector rotCamOff, destPos;
			angle.Rotate(&chaseOffset, &rotCamOff);
			destPos = *trackPosition;
			destPos.add(&rotCamOff);

			//LERP from the current track position to the destination track position
			//btw, how 'bout that convergance of geometric infinite series?

			//calculate the change in position from current pos to track pos
			Vector del(&destPos);
			del.sub(&position);
			//scale it by our convergance coefficient
			del.scale(trackLerpC);
			//and add it to the last position
			position.add(&del);
		}

		break;
	}
}
