#include "collision.h"
#include "entity.h"
#include "vector.h"
#include "geometry.h"
#include "main.h"
#include "const.h"

#include <windows.h>	//NULL

#define C_MODEL_COUNT		3

#define C_MODEL_OBJ0_STATE0	0
#define C_MODEL_OBJ0_STATE1	1
#define C_MODEL_OBJ1		2

#define C_MODEL_VERTEX_MAX	100
#define C_MODEL_PLANE_MAX	100


int		cmodel_vtxCount[C_MODEL_COUNT];
Vector	cmodel_vtx[C_MODEL_COUNT][C_MODEL_VERTEX_MAX];	//2400 bytes of collision verticies
int		cmodel_planeCount[C_MODEL_COUNT];
Plane	cmodel_plane[C_MODEL_COUNT][C_MODEL_PLANE_MAX];	//4800 bytes of collision planes
int		cmodel_edgeCount[C_MODEL_COUNT];
int		*cmodel_edge[C_MODEL_COUNT];

//oldschool - calc_bbox compatibility - for those entities with no collision model
static int cube_edge[24] = {0,1,1,3,3,2,2,0,4,5,5,7,7,6,6,4,0,4,1,5,2,6,3,7};

void LoadCollisionModel(CModel *c_model, Vector *position, Quat *angle, float scale, int cid)
{
	int v;

	//make sure its a valid collision model ID
	if (cid < 0 || cid >= C_MODEL_COUNT) return;

	//if no collision model data - bail out on the extensive tests
	if (!c_model) return;

	cmodel_vtxCount[cid] = c_model->vtxCount;
	if (cmodel_vtxCount[cid] > C_MODEL_VERTEX_MAX)
	{
		main_console->Print("Encountered collision model %s with vertex count %d greater than vertex buffer\n", c_model->title, c_model->vtxCount);
		return;
	}

	cmodel_planeCount[cid] = c_model->planeCount;
	if (cmodel_planeCount[cid] > C_MODEL_PLANE_MAX)
	{
		main_console->Print("Encountered collision model %s with plane count %d greater than plane buffer\n", c_model->title, c_model->planeCount);
		return;
	}

	cmodel_edgeCount[cid] = c_model->edgeCount;
	cmodel_edge[cid] = c_model->edge;

	//calculate verticies
	for (v = 0; v < cmodel_vtxCount[cid]; v++)
	{
		//scale
		cmodel_vtx[cid][v].scale(&c_model->vtx[v], scale);
		//rotate
		angle->Rotate(&cmodel_vtx[cid][v]);
		//translate
		cmodel_vtx[cid][v].add(position);
	}

	//calculate planes... (this is gonna be nasty)
	for (v = 0; v < cmodel_planeCount[cid]; v++)
	{
		//scale
		cmodel_plane[cid][v].pos.scale(&c_model->plane[v].pos, scale);
		//rotate
		angle->Rotate(&cmodel_plane[cid][v].pos);
		angle->Rotate(&c_model->plane[v].normal, &cmodel_plane[cid][v].normal);
		//translate
		cmodel_plane[cid][v].pos.add(position);
	}

	//copy over debug stuff - invalid - i just elimitaed "Entity *e"
#ifdef DEBUG_CMODEL
	e->d_cmodel_vtxCount = cmodel_vtxCount[cid];
	memcpy(e->d_cmodel_vtx, cmodel_vtx[cid], sizeof(Vector)*cmodel_vtxCount[cid]);
	e->d_cmodel_planeCount = cmodel_planeCount[cid];
	memcpy(e->d_cmodel_plane, cmodel_plane[cid], sizeof(Plane)*cmodel_planeCount[cid]);
	e->d_cmodel_edgeCount = cmodel_edgeCount[cid];
	e->d_cmodel_edge = cmodel_edge[cid];
#endif
}



/*
PlaneSeperationState

	searches for a dividing plane from the p list
	that contains all verticies of the v list on one side of it
	
	returns
		0 = if any plane excludes all verticies
		1 = if all planes contain all verticies
		2 = if, for any plane, some points lie on one side and some on others

*/
int PlaneSeperationState(Plane *p, int pcount, Vector *v, int vcount)
{
	int i,j;
	bool outside, inside;

	//only if all verticies pass all plane tests will this return true
	inside = true;

	//check all camera frustum planes
	for (i = 0; i < pcount; i ++)
	{
		//reset this when checking every plane
		outside = true;

		for (j = 0; j < vcount; j++)
		{
			float distance_from_plane = p[i].Distance(&v[j]);

			outside &= distance_from_plane > 0;
			inside &= distance_from_plane < 0;
		}

		//if all points lie on the outside of this plane
		//then plane[i] will be true
		//and the object will be to be clipped
		if (outside) return 0;
	}

	if (inside) return 1;
	return 2;
}



/*

FindDividingPlane	- mutation of PlaneSeperationState

	searches for a dividing plane from the p list
	that contains all verticies of the v list on one side of it
	
	if a plane has been found that holds all verticies on its positive side
		return the ID of that plane
	else - if no plane holds any verticies
		return -1

*/
int FindDividingPlane(Plane *p, int pcount, Vector *v, int vcount)
{
	int i,j;
	bool outside;

	//check all camera frustum planes
	for (i = 0; i < pcount; i ++)
	{
		//reset this when checking every plane
		outside = true;

		for (j = 0; j < vcount; j++)
		{
			int ja = j>>5;	//the dword
			int jb = j&31;	//the bit

			float distance_from_plane = p[i].Distance(&v[j]);

			//if dist_from_plane is outside then maintain 'outside' status
			//if not - we know its not outside - and we can bail out
			outside &= distance_from_plane > 0;
			if (!outside) break;
		}

		//if all points lie on the outside of this plane
		//then plane[i] will be true
		//and the object will be to be clipped
		if (outside) return i;
	}
	return -1;
}


/*
ObjectCollision

pass to this function:
two objects (object1 and object2) each containing:
	a plane array / plane array count
	a vertex array / vertex array count
	an edge array / edge array count

plane array consists of v_count consecutive planes 
vertex array consists of p_count consecutive vectors
edge array consists of e_count*2 consecutive integers
	each pair of integers specifies two indicies of verticies within the vertex array

first - check all surface planes of one object to all verticies of another
	if a seperating plane exists then return 0
next - check all planes composed by all edges of one object and all verticies of another
	if a dividing plane 
*/
int ObjectCollision(void)
{
	Plane *p1 =		cmodel_plane[C_MODEL_OBJ0_STATE0];
	int p1count =	cmodel_planeCount[C_MODEL_OBJ0_STATE0];
	Vector *v1 =	cmodel_vtx[C_MODEL_OBJ0_STATE0];
	int v1count =	cmodel_vtxCount[C_MODEL_OBJ0_STATE0];
	int *e1 =		cmodel_edge[C_MODEL_OBJ0_STATE0];
	int e1count =	cmodel_edgeCount[C_MODEL_OBJ0_STATE0];

	Plane *p2 =		cmodel_plane[C_MODEL_OBJ1];
	int p2count =	cmodel_planeCount[C_MODEL_OBJ1];
	Vector *v2 =	cmodel_vtx[C_MODEL_OBJ1];
	int v2count =	cmodel_vtxCount[C_MODEL_OBJ1];
	int *e2 =		cmodel_edge[C_MODEL_OBJ1];
	int e2count =	cmodel_edgeCount[C_MODEL_OBJ1];
	
	//for all planes of the bounding box e
		//for all planes of the bounding box main_ent
			//if any one does not seperate all verticies of the other

	//-1 return implies no seperating plane
	if (FindDividingPlane(p1, p1count, v2, v2count) > -1) return 0;
	if (FindDividingPlane(p2, p2count, v1, v1count) > -1) return 0;

	//now that we've checked all planes, check all edge/vertex pairs - check for the lowest edge count?
	Plane plane;
	int side1, side2;		//1 - front, 0 - back
	float distance_from_plane;
	bool intersection;

	//check all edges of object 1 ...
	for (int e = 0; e < e1count*2; e+=2)
	{
		//... to all verticies of object 2
		for (int i = 0; i < v2count; i++)
		{
			//set intersection to 'not yet'
			intersection = false;

			//set both object side-of-planes to 'undefined'
			side1 = side2 = -1;

			//set the normal based on the edge and the vertex
			plane.normal.CalcNormal(&v1[e1[e]], &v1[e1[e+1]], &v2[i]);
			plane.pos = v2[i];

			//check all verticies of object 1
			for (int v = 0; v < v1count; v++)
			{
				//dont check any points on the plane
				if (v == e1[e]) continue;
				if (v == e1[e+1]) continue;

				//grab current distance from plane to checked vertex
				float distance_from_plane = plane.Distance(&v1[v]);

				//if a side hasnt been set, then set it
				if (side1 == -1) side1 = distance_from_plane > 0;
				//else, check the current vertex to the side, and if it differs - we have intersection with this plane
				else if (side1 != (distance_from_plane > 0))
				{
					intersection = true;
					break;
				}
			}

			if (intersection) continue;

			//if a side has been defined for object 1 then set object 2's side to the opposite of object 1
			if (side1 != -1) side2 = !side1;

			//check all verticies of object 2
			for (int v = 0; v < v2count; v++)
			{
				if (v == i) continue;

				distance_from_plane = plane.Distance(&v2[v]);

				if (side2 == -1) side2 = distance_from_plane > 0;
				else if (side2 != (distance_from_plane > 0))
				{
					intersection = true;
					break;
				}
			}

			if (intersection) continue;

			//the planes has been checked and found to be a dividing plane
			return 0;
		}
	}

	//all planes have been checked, and all divided.  return "intersection occured"
	return 1;
}

// ****************************************
//			COLLISION TREE MODEL
// ****************************************
#define COL_TREE_CHILDREN	8
Entity **ent_tree_list = NULL;
typedef struct col_tree_s {
	struct col_tree_s	*child[COL_TREE_CHILDREN];			//the node children
	Box					range;							//the range of the node
	Entity				**entList;					//the entity pointer list of the node
	int					entListCount;			//number of total entity pointers
	int					entNodeListCount;	//number of front entity pointers that only lie within the parent node
} col_tree_t;

#define COL_tree_MAX	256
col_tree_t col_tree[COL_tree_MAX];
int col_tree_next;
int col_tree_max_depth;

int ColTreeMaxEnts = 3;
int e_ShrinkTreeBBox = 1;

void RecurseCreateCollisiontree(col_tree_t *parent, int depth)
{
	int v;
	Vector del;
	Vector mid;

//check max depth
	if (col_tree_max_depth < depth) col_tree_max_depth = depth;

//check to see if this box holds too few children to subdivide
	if (parent->entListCount <= ColTreeMaxEnts) return;

	mid = parent->range.Center();

//subdivide the space
	Box sp[COL_TREE_CHILDREN];	//0 - min, 1 - max
	//for each octant
	for (v = 0; v < COL_TREE_CHILDREN; v++)
	{
		//for each bit vector
		for (int b = 0; b < 3; b++)
		{
			if (!(v & (1 << b)))	//minimum x
			{
				sp[v].min.v[b] = parent->range.min.v[b];
				sp[v].max.v[b] = mid.v[b];
			}
			else
			{
				sp[v].min.v[b] = mid.v[b];
				sp[v].max.v[b] = parent->range.max.v[b];
			}
		}
	}

	//sort by child - swap from back to front
	//quicker sorting method?
	int next_swap = parent->entListCount-1;
	for (int c = 0; c < COL_TREE_CHILDREN; c++)
	{
		for (v = parent->entListCount-1; v >= 0; v--)
		{
			//only those in the list which are within the current spacial divide sp[c]
			if (!sp[c].Contains(&parent->entList[v]->maxbbox)) continue;

			//swap 'em
			Entity *e_swap;
			e_swap = parent->entList[v];
			parent->entList[v] = parent->entList[next_swap];
			parent->entList[next_swap] = e_swap;

			//if the child doesnt exist - and theres room to create a new child - then create it
			if (!parent->child[c] && col_tree_next < COL_tree_MAX)
			{
				//grab the new child
				parent->child[c] = &col_tree[col_tree_next];
				col_tree_next++;
				//set its pointer / counter / space
				if (e_ShrinkTreeBBox)
				{
					parent->child[c]->range.max = parent->child[c]->range.min = vectorInf;
					parent->child[c]->range.max.negative();
				}
				else
				{
					parent->child[c]->range = sp[c];
				}
				//zero all else
				memset(parent->child[c]->child, 0, sizeof(Entity **) * COL_TREE_CHILDREN);
				parent->child[c]->entListCount = 0;
				parent->child[c]->entNodeListCount = 0;
			}
			if (parent->child[c])
			{
				//reset the child node list pointer
				parent->child[c]->entList = &parent->entList[next_swap];
				//up its counter and its overall node counter
				parent->child[c]->entListCount++;
				parent->child[c]->entNodeListCount++;
				if (e_ShrinkTreeBBox) 
					parent->child[c]->range.Stretch(&parent->entList[next_swap]->maxbbox);
				//decrease the parent counter
				parent->entNodeListCount--;
			}
			next_swap--;
		}
	}

	depth++;
	for (int c = 0; c < COL_TREE_CHILDREN; c++)
		if (parent->child[c])
			RecurseCreateCollisiontree(parent->child[c], depth);
}

void Entity_ParseCollisiontree(void)
{
	int a;

	//allocate the array if its not there quite yet
	if (!ent_tree_list) ent_tree_list = (Entity **)malloc(MAX_ENTS * sizeof(Entity *));
	if (!ent_tree_list) return;

	//create base node
	memset(col_tree->child, 0, sizeof(Entity **) * COL_TREE_CHILDREN);
//	col_tree->range.min = col_tree->range.max = vectorInf;
//	col_tree->range.max.negative();
	col_tree->range.min = Vector(-WorldSize*2,-WorldSize*2,-WorldSize*2);
	col_tree->range.max = Vector(WorldSize*2,WorldSize*2,WorldSize*2);
	col_tree->entList = ent_tree_list;
	col_tree->entListCount = 0;
	for (a = 0; a < MAX_ENTS; a++)
	{
		if (entityarray[a].active &&
			entityarray[a].check_collision &&
			entityarray[a].c_model)
		{
			//place all entities within the root node pointer
			col_tree->entList[col_tree->entListCount] = &entityarray[a];
			//up the root node maximum count
			col_tree->entListCount++;
			//and stretch the root node bounding box
//			col_tree->range.Stretch(&entityarray[a].maxbbox);
		}
	}
	col_tree->entNodeListCount = col_tree->entListCount;
	col_tree_next = 1;

	//subdivide and sort
	col_tree_max_depth = 0;
	RecurseCreateCollisiontree(col_tree, 0);
}

// ****************************************
//				COLLISION UPDATE
// ****************************************

int e_TestHull = 1;
int e_TestSphere = 1;
int e_TestBBox = 1;

/*
the only objects linked to the collision tree
	*have 'active'
	*have 'check_collision'
	*have 'c_model'
*/
void Entity_CheckCollisiontree(col_tree_t *parent)
{
	//new state
	Box new_maxbbox;
	Vector new_position;
	Quat new_angle;
	//collision info
	Vector intersection;
	float movement_fraction;

	//indicies
	int a,b;
	bool main_ent_loaded;
	bool collide_overall, collide_ent;

	//check all entities within the current node
	for (a = 0; a < parent->entNodeListCount; a++)
	{
		main_ent = parent->entList[a];

		//calc new entity position
		new_position.add(&main_ent->position, &main_ent->velocity);
		new_maxbbox.min.add(&main_ent->maxbbox.min, &main_ent->velocity);
		new_maxbbox.max.add(&main_ent->maxbbox.max, &main_ent->velocity);

		//update rotation
		new_angle = main_ent->angle;
		if (main_ent->rotation.w != 1)
		{
			new_angle.MulTo(&main_ent->rotation);
			new_angle.Normalize();
		}

		movement_fraction = 1.f;

		//set this object to have not been collided quite yet
		collide_overall = false;

		//tell the collision check code that we have not yet loaded this entity's info
		main_ent_loaded = false;

		//check main_ent to all other entities over node 'n node children
		for (b = a + 1; b < parent->entListCount; b++)
		{
			Entity *e = parent->entList[b];

			//assume collision until a test fails
			collide_ent = true;

			//Axis-Aligned Bounding Box Test
			//if their maximum bounding boxes dont intersect - then forget testing
			if (e_TestBBox)
				if (!new_maxbbox.Intersects(&e->maxbbox))
					collide_ent = false;

			//Sphere Test
			//if the distance between the two entities is greater than the summed distance between 
			//the collision models - then forget about it
			if (e_TestSphere && collide_ent)
			{
				Vector del;
				float nrml;
				del.sub(&e->position, &new_position);
				nrml = e->c_model->radius*e->scale + main_ent->c_model->radius*main_ent->scale;
				nrml *= nrml;
				if (del.normal() > nrml)
					collide_ent = false;
			}

			//And the true to life Convex Hull test
			if (e_TestHull && collide_ent)
			{
				if (!main_ent_loaded)
				{
					LoadCollisionModel(
						main_ent->c_model,
						&new_position,
						&new_angle,
						main_ent->scale,
						C_MODEL_OBJ0_STATE0);

					main_ent_loaded = true;
				}
				
				LoadCollisionModel(
					e->c_model,
					&e->position,
					&e->angle,
					e->scale,
					C_MODEL_OBJ1);

				if (!ObjectCollision())
					collide_ent = false;
			}

			if (collide_ent)
			{
				//if there is no dividing plane for either of the loaded collision models
				//ie the objects *have* touched
				main_ent->touched = true;
				e->touched = true;

				collide_overall = true;
//				new_position = main_ent->position;
//				new_angle = main_ent->angle;

				if (main_ent->Touch)
				{
					touch_ent = e;
					main_ent->Touch();
				}
				//main_ent died - stop checking for more collisions
				if (!EntIsActive(main_ent)) break;
				//e died - dont do the other check
				if (EntIsActive(e))
				{
					if (e->Touch)
					{
						touch_ent = main_ent;
						main_ent = e;
						e->Touch();
						main_ent = touch_ent;
					}
				}
				//main_ent died - stop checking for more collisions
				if (!EntIsActive(main_ent)) break;
			}
		}

		if (EntIsActive(main_ent))
		{
			main_ent->position = new_position;
			main_ent->angle = new_angle;
		}
	}

	//recursively check children - when they exist
	for (a = 0; a < COL_TREE_CHILDREN; a++)
		if (parent->child[a])
			Entity_CheckCollisiontree(parent->child[a]);
}

void Ent_AlignBBox(Entity *e)
{
	Vector vtx[8];
	e->bbox.CalcVtxs(vtx);		//the eight verticies
	//rotate/scale/translate and stretch max bbox

	e->maxbbox.min = Vector(fInf,fInf,fInf);
	e->maxbbox.max = Vector(-fInf,-fInf,-fInf);
	for (int v = 0; v < 8; v++)
	{
		//scale
		vtx[v].scale(e->scale);
		//rotate
		e->angle.Rotate(&vtx[v]);
		//translate
		vtx[v].add(&e->position);
		//and stretch
		e->maxbbox.Stretch(&vtx[v]);
	}
}

void Entity_UpdateCollision(void)
{
	static bool init_d = false;
	if (!init_d)
	{
		init_d = true;
		new M_CVar(&ColTreeMaxEnts,		"ColTreeMaxEnts");
		new M_CVar(&col_tree_next,		"ColTreeCount");
		new M_CVar(&e_TestHull,			"TestHull");
		new M_CVar(&e_TestSphere,		"TestSphere");
		new M_CVar(&e_TestBBox,			"TestBBox");
		new M_CVar(&e_ShrinkTreeBBox,	"ShrinkTreeBBox");
		new M_CVar(&col_tree_max_depth, "ColTreeMaxDepth");
	}

		//physics collision tree requires all maximum bbox's to be precalculated
		//update max bbox
	for (int a = 0; a < MAX_ENTS-1; a++)
		if (entityarray[a].active && entityarray[a].check_collision && entityarray[a].c_model)
			Ent_AlignBBox(&entityarray[a]);

	//seperate all entities into a tree bin - only place those fully within the child range inside the child node
	Entity_ParseCollisiontree();

	//check them recursively, starting with the first
	if (col_tree) Entity_CheckCollisiontree(col_tree);

	//then cycle through all non-colliding objects, and update them on their own
	for (int a = 0; a < MAX_ENTS; a++)
	{
		main_ent = &entityarray[a];
		if (main_ent->active && (!main_ent->check_collision || !main_ent->c_model))
		{
			main_ent->position.add(&main_ent->velocity);
			if (main_ent->rotation.w != 1)
			{
				main_ent->angle.MulTo(&main_ent->rotation);
				main_ent->angle.Normalize();
			}
		}
	}
}