#include <windows.h>
#include <gl/gl.h>

#include "entity.h"
#include "random.h"
#include "error.h"

#include "main.h"

Entity *main_ent = NULL;
Entity *touch_ent = NULL;
Entity *client_ent = NULL;
Entity *entityarray = NULL;

bool EntIsActive(Entity *e)
{
	return (e && e->active);
}

// ********** CONSTRUCTION/DESTRUCTION ********** 

void Entity::Init(void)
{
	active = true;

	check_collision = true;

	position = Vector(0,0,0);
	velocity = Vector(0,0,0);
	angle = Quat(0,0,0,1);
	rotation = Quat(0,0,0,1);

	bbox.min = Vector(0,0,0);
	bbox.max = Vector(0,0,0);

	texture = NULL;
	model = NULL;
	c_model = NULL;

	color.l = 0xFFFFFFFF;
	scale = 1.f;

	Think = NULL;
	Touch = NULL;
}

void Entity::Kill(void)
{
	active = false;

	if (this == main_ent) main_ent = NULL;
	if (this == touch_ent) touch_ent = NULL;
	if (this == client_ent) client_ent = NULL;
	//unlink this entity from all other object pointers (including cameras, etc) *HERE*
	//not from the entity list, however. that will have been already done by the chain destructor

	//models and textures are not to be unloaded also.  they are not loaded within the entity code,
	//instead pointers are assigned by the external map object creation


}

//********* ENTITY FUNCTIONS **********  called by Engine


#include "console.h"

int BBoxDebugDisplay = 0;
int EntityDisplaySoftware = 0;
int e_DisplayEntities = 1;
int e_Invincible = 1;
int e_UpdateCollision = 1;

static M_CVar *n[] = {
	new M_CVar(&BBoxDebugDisplay,		"BBoxDebugDisplay"),
	new M_CVar(&EntityDisplaySoftware,	"EntityDisplaySoftware"),
	new M_CVar(&e_DisplayEntities,		"DisplayEntities"),
	new M_CVar(&e_Invincible,			"Invincible"),
	new M_CVar(&e_UpdateCollision,		"UpdateCollision"),
	NULL};

Vector temp_vtx[250];
Vector temp_normal[250];


void Entity::Display(void)
{
	if (EntityDisplaySoftware)
	{

		int v;

		//debug display main bounding box
		if (BBoxDebugDisplay)
		{
			Vector bbvtx[8];
			maxbbox.CalcVtxs(bbvtx);
			//debug - display maximum bounding box
			static int cube_lines[24] = {0,1,1,3,3,2,2,0,4,5,5,7,7,6,6,4,0,4,1,5,2,6,3,7};
			glBindTexture(GL_TEXTURE_2D, 0);
			glColor3f(touched,!touched,0);	//red if touched, green if not
			glBegin(GL_LINES);
			for (int v = 0; v < 24; v++) glVertex3fv(bbvtx[cube_lines[v]].v);
			glEnd();
		}

		//debug display collision model
#ifdef DEBUG_CMODEL
		if (BBoxDebugDisplay)
		{
			int e;
			glBindTexture(GL_TEXTURE_2D, 0);
			glBegin(GL_LINES);
			glColor3f(1,1,touched);	//red if touched, green if not
			for (e = 0; e < d_cmodel_edgeCount * 2; e++)
				glVertex3fv(d_cmodel_vtx[d_cmodel_edge[e]].v);
			glColor3f(1,0.5f,touched);
			for (e = 0; e < d_cmodel_planeCount; e++)
			{
				Vector n_sc;
				n_sc.scale(&d_cmodel_plane[e].normal, c_model->radius);
				n_sc.add(&d_cmodel_plane[e].pos);
				glVertex3fv(d_cmodel_plane[e].pos.v);
				glVertex3fv(n_sc.v);
			}
			glEnd();
		}
#endif

		if (BBoxDebugDisplay)
		{
			//debug - display axii
			//our debug world coordinates
			glBegin(GL_LINES);
			for (v = 0; v < 3; v++)
			{
				glColor3f(base_axis[v].x, base_axis[v].y, base_axis[v].z);
				glVertex3fv(position.v);
				glVertex3f(	position.x+axis[v].x*c_model->radius*2,
							position.y+axis[v].y*c_model->radius*2,
							position.z+axis[v].z*c_model->radius*2);
			}
			glEnd();
		}

		for (v = 0; v < model->vtxCount; v++)
		{
			angle.Rotate(&model->vtx[v], &temp_vtx[v]);
			angle.Rotate(&model->normal[v], &temp_normal[v]);
			if (scale != 1.f) temp_vtx[v].scale(scale);
			temp_vtx[v].add(&position);
		}

		//set the texture
		SetTexture(texture);

		//set the color
		glColor4ubv(color.rgba);

		Model m;

		//align the newly created model
		m.vtxCount =	model->vtxCount;
		m.vtx =			temp_vtx;
		m.normal =		temp_normal;
		m.texCoord =	model->texCoord;
		m.triCount =	model->triCount;
		m.tri =			model->tri;

		if (transparent)
		{
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glEnable(GL_BLEND);
		}

		//draw the model
		DisplayModel(&m);

		if (transparent) glDisable(GL_BLEND);

		//NULL everything before we exit the function - so the destructor doesnt do any unnecesssary 'free'ing
		m.vtx = NULL;
		m.normal = NULL;
		m.texCoord = NULL;
		m.tri = NULL;
	}
	else
	{
		int v;
		glPushMatrix();

		//debug display main bounding box
		if (BBoxDebugDisplay)
		{
			Vector bbvtx[8];
			maxbbox.CalcVtxs(bbvtx);
			//debug - display maximum bounding box
			static int cube_lines[24] = {0,1,1,3,3,2,2,0,4,5,5,7,7,6,6,4,0,4,1,5,2,6,3,7};
			glBindTexture(GL_TEXTURE_2D, 0);
			glColor3f(touched,!touched,0);	//red if touched, green if not
			glBegin(GL_LINES);
			for (int v = 0; v < 24; v++) glVertex3fv(bbvtx[cube_lines[v]].v);
			glEnd();
		}

		//debug display collision model
#ifdef DEBUG_CMODEL
		if (BBoxDebugDisplay)
		{
			int e;
			glBindTexture(GL_TEXTURE_2D, 0);
			glBegin(GL_LINES);
			glColor3f(1,1,touched);	//red if touched, green if not
			for (e = 0; e < d_cmodel_edgeCount * 2; e++)
				glVertex3fv(d_cmodel_vtx[d_cmodel_edge[e]].v);
			glColor3f(1,0.5f,touched);
			for (e = 0; e < d_cmodel_planeCount; e++)
			{
				Vector n_sc;
				n_sc.scale(&d_cmodel_plane[e].normal, c_model->radius);
				n_sc.add(&d_cmodel_plane[e].pos);
				glVertex3fv(d_cmodel_plane[e].pos.v);
				glVertex3fv(n_sc.v);
			}
			glEnd();
		}
#endif

		//offset the model
		glTranslatef(position.x, position.y, position.z);

		if (BBoxDebugDisplay)
		{
			//debug - display axii
			//our debug world coordinates
			glPushMatrix();
			glScalef(2*c_model->radius,2*c_model->radius,2*c_model->radius);
			glBegin(GL_LINES);
			for (v = 0; v < 3; v++)
			{
				glColor3fv(base_axis[v].v);
				glVertex3f(0,0,0);
				glVertex3fv(axis[v].v);
			}
			glEnd();
			glPopMatrix();
		}

		Quat q;
		q.QuatToAngleAxis(&angle);
		//rotate it
		glRotatef(q.w, q.x, q.y, q.z);

		//set the texture
		SetTexture(texture);

		//set the color
		glColor4ubv(color.rgba);

		//scale the model
		if (scale != 1.f)
		{
			glScalef(scale, scale, scale);
			glEnable(GL_NORMALIZE);
		}

		if (transparent)
		{
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glEnable(GL_BLEND);
		}

		//draw the model
		DisplayModel(model);

		if (transparent) glDisable(GL_BLEND);

		glDisable(GL_NORMALIZE);
		glPopMatrix();

	}

}




//CHEAP-ASS-HACK-ENTITY-SYSTEM

bool InitEntities(void)
{
	entityarray = (Entity *)malloc(MAX_ENTS * sizeof(Entity));
	if (!entityarray) return false;
	memset(entityarray, 0, MAX_ENTS * sizeof(Entity));
	return true;
}

void ShutdownEntities(void)
{
	if (entityarray) free(entityarray);
}

Entity *CreateEntity(void)
{
	for (int v = 0; v < MAX_ENTS; v++)
	{
		if (!entityarray[v].active)
		{
			entityarray[v].Init();
			return &entityarray[v];
		}
	}
	return NULL;
}

void RemoveEntity(Entity *e)
{
	e->Kill();
}

void DisplayEntities(void)
{
	if (e_DisplayEntities)
		for (int v = 0; v < MAX_ENTS; v++)
			if (entityarray[v].active)
				entityarray[v].Display();
}



/* SHOT FUNCTIONS */

#include <math.h>

void Fade_Think(void)
{
	main_ent->timer--;

	float x = 1.f - (float)main_ent->timer / 60.f;

	main_ent->color.rgba[3] = (int)(127.f * sqrt(1.f - x*x));
	if (main_ent->timer < 0) RemoveEntity(main_ent);
}

void Rock_Think(void)
{
	main_ent->timer--;

	main_ent->transparent = true;
	main_ent->color.rgba[3] = 0x7F;

	float size = MCRand();
	Vector vel(MCRand(),MCRand(),MCRand());
	vel.scale(4.f*(1-size));
	main_engine->particleSystem->CreateParticle(&main_ent->position, &vel, 30*size+1, 0xFF0000FF, MRand()%50);
	
	if (main_ent->timer < 0)
	{
		main_ent->Think = Fade_Think;
		main_ent->timer = 60;
	}
}

void Shot_Touch(void)
{
	if (!touch_ent) return;

	if (!strcmp(touch_ent->type, "player")) return;

	int v;
	float rad;

	if (!strcmp(touch_ent->type, "rock_l"))
	{
		touch_ent->check_collision = false;
		for (v = 0; v < 4; v++)
		{
			Entity *e = CreateEntity();
			if (!e) return ;

			e->model = GetModel(&main_engine->modelChain, "WorldData/Models/rock.mdl");
			e->c_model = GetCModel(&main_engine->cmodelChain, "WorldData/Models/c_rock.mdl");
			e->texture = GetTexture(&main_engine->skinChain, "WorldData/Skins/rock.bmp");

			e->color.l = 0xFF7F7F7F;

			e->velocity = Vector(MCRand(), MCRand(), MCRand());
			e->rotation.x = MCRand();
			e->rotation.y = MCRand();
			e->rotation.z = MCRand();
			e->rotation.w = 0;
			e->rotation.Normalize();
			e->rotation.w = MCRand();
			e->rotation.AngleAxisToQuat();
			strcpy(e->type, "rock_m");

			e->scale = touch_ent->scale * 0.5f;
			e->bbox = e->c_model->bbox;

			e->velocity.scale(3);

			e->position = Vector(MCRand(), MCRand(), MCRand());
			e->position.normalize();
			e->position.scale(e->c_model->radius * e->scale);
			e->position.add(&touch_ent->position);
		}
	}
	else if (!strcmp(touch_ent->type, "rock_m"))
	{
		touch_ent->check_collision = false;
		for (v = 0; v < 4; v++)
		{
			Entity *e = CreateEntity();
			if (!e) return ;

			e->model = GetModel(&main_engine->modelChain, "WorldData/Models/rock.mdl");
			e->c_model = GetCModel(&main_engine->cmodelChain, "WorldData/Models/c_rock.mdl");
			e->texture = GetTexture(&main_engine->skinChain, "WorldData/Skins/rock.bmp");

			e->color.l = 0xFF003F7F;

			e->velocity = Vector(MCRand(), MCRand(), MCRand());
			e->rotation.x = MCRand();
			e->rotation.y = MCRand();
			e->rotation.z = MCRand();
			e->rotation.w = 0;
			e->rotation.Normalize();
			e->rotation.w = MCRand();
			e->rotation.AngleAxisToQuat();
			e->bbox = e->c_model->bbox;
			strcpy(e->type, "rock_s");

			e->scale = touch_ent->scale * 0.5f;

			e->velocity.scale(10);

			e->position = Vector(MCRand(), MCRand(), MCRand());
			e->position.normalize();
			e->position.scale(e->c_model->radius * e->scale);
			e->position.add(&touch_ent->position);
		}
	}
	else if (!strcmp(touch_ent->type, "rock_s"))
	{
		touch_ent->check_collision = false;
		for (int t = 0; t < 30; t++)
		{
			Entity *e = CreateEntity();
			if (!e) return ;

			e->model = GetModel(&main_engine->modelChain, "WorldData/Models/rock.mdl");
			e->c_model = GetCModel(&main_engine->cmodelChain, "WorldData/Models/c_rock.mdl");
			e->texture = GetTexture(&main_engine->skinChain, "WorldData/Skins/rock.bmp");

			e->angle = Quat(MCRand(),MCRand(),MCRand(),MCRand());
			e->angle.Normalize();
			e->rotation.x = MCRand();
			e->rotation.y = MCRand();
			e->rotation.z = MCRand();
			e->rotation.w = 0;
			e->rotation.Normalize();
			e->rotation.w = MCRand();
			e->rotation.AngleAxisToQuat();
			e->bbox = e->c_model->bbox;
			strcpy(e->type, "dust");

			e->timer = 100 + MRand() % 100;
			e->Think = Rock_Think;

			e->color.l = 0xFF0000CF;

			e->check_collision = false;

			rad = MFRand();

			e->scale = touch_ent->scale * .3f * (1.25f - rad);

			for (v = 0; v < 3; v++) e->position.v[v] = MCRand();
			e->position.normalize();
			for (v = 0; v < 3; v++)  e->velocity.v[v] = e->position.v[v] + touch_ent->velocity.v[v] * 0.25f;
			e->velocity.normalize();
			for (v = 0; v < 3; v++) e->position.v[v] = e->position.v[v] * (touch_ent->c_model->radius * touch_ent->scale * 0.5f * rad) + touch_ent->position.v[v];
			for (v = 0; v < 3; v++)	e->velocity.v[v] = (e->velocity.v[v] * 4.f + touch_ent->velocity.v[v] * .75f) * (.25f + rad);
		}
	}

	for (int tvar = 0; tvar < 50; tvar++)
	{
		rad = MFRand()*MFRand() + MFRand();
		Vector vel(MCRand(),MCRand(),MCRand());
		vel.normalize();
		vel.scale(7.f*(rad-.75f));
		color_t color;
		color.rgba[0] = 0xFF;
		color.rgba[1] = MRand() & 0x7F;
		color.rgba[2] = color.rgba[1] >> 1;
		color.rgba[3] = 0xFF;
		main_engine->particleSystem->CreateParticle(&main_ent->position, &vel, 100.f*(2.25f - rad), color.l, 50 + (MRand() % 50));
	}

	RemoveEntity(touch_ent);
	RemoveEntity(main_ent);
}

void Shot_Think(void)
{
	Vector vel,vy,vz;

	static int timer = 0;
	timer++;
	if (timer % 5 == 0)
	{
		vz.scale(&main_ent->axis[0], -MFRand());
		vy.scale(&main_ent->axis[1], MCRand());
		vel.scale(&main_ent->axis[2], MCRand());
		vel.add(&vy);
		vel.add(&vz);

		main_engine->particleSystem->SetBlend(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		main_engine->particleSystem->CreateParticle(&main_ent->position, &vel, 50.f + 150.f * MFRand(), 0x7FFFFFFF, MRand()%150 + 150);
		main_engine->particleSystem->SetBlend(GL_SRC_ALPHA, GL_ONE);
	}

	main_ent->timer--;
	if (main_ent->timer < 0) RemoveEntity(main_ent);
}

void Ent_FireShot(Entity *e)
{
	if (e->timer > M_Engine::Time) return;

	e->timer = M_Engine::Time + 1.f;

	Entity *n = CreateEntity();
	if (!n) return;

	n->model = GetModel(&main_engine->modelChain, "WorldData/Models/shot.mdl");
	n->c_model = GetCModel(&main_engine->cmodelChain, "WorldData/Models/c_shot.mdl");
	n->texture = GetTexture(&main_engine->skinChain, "WorldData/Skins/shot.bmp");

	n->angle = e->angle;

	n->position = e->axis[0];
	n->position.scale(100);
	n->position.add(&e->position);
	n->position.add(&e->velocity);

	n->velocity = Vector(25,0,0);
	n->angle.Rotate(&n->velocity);

	n->rotation = Quat(1,0,0,5);
	n->rotation.AngleAxisToQuat();
	
	n->timer = 250;
	n->Think = Shot_Think;
	n->Touch = Shot_Touch;

	n->bbox = n->c_model->bbox;	//setsize

	strcpy(n->type, "shot");
}

/* PLAYER FUNCTIONS */

void Player_Think(void)
{
	//debug - decay velocity
	main_ent->velocity.scale(0.9975f);
}

void Player_Touch(void)
{
	if (e_Invincible) return;

	if (strcmp(touch_ent->type, "rock_s") &&
		strcmp(touch_ent->type, "rock_m") &&
		strcmp(touch_ent->type, "rock_l"))
		return;

	// DIE!!!

	client_ent = NULL;

	for (int v = 0; v < 20; v++)
	{
		Entity *e = CreateEntity();
		if (!e) return ;

		e->model = GetModel(&main_engine->modelChain, "WorldData/Models/rock.mdl");
		e->c_model = GetCModel(&main_engine->cmodelChain, "WorldData/Models/c_rock.mdl");
		e->texture = GetTexture(&main_engine->skinChain, "WorldData/Skins/ship.bmp");

		e->angle = Quat(MCRand(),MCRand(),MCRand(),MCRand());
		e->angle.Normalize();
		e->rotation.x = MCRand();
		e->rotation.y = MCRand();
		e->rotation.z = MCRand();
		e->rotation.w = 0;
		e->rotation.Normalize();
		e->rotation.w = MCRand();
		e->rotation.AngleAxisToQuat();
		e->bbox = e->c_model->bbox;
		strcpy(e->type, "dust");

		e->velocity = Vector(MCRand(), MCRand(), MCRand());
		e->velocity.scale(3);
		e->velocity.add(&main_ent->velocity);
		e->timer = 100;
		e->Think = Rock_Think;

		e->scale = .1f;

		e->color.l = 0xFF00004F;

		e->check_collision = false;

		e->position = Vector(MCRand(), MCRand(), MCRand());
		e->position.normalize();
		e->position.scale(e->c_model->radius * main_ent->scale);
		e->position.add(&main_ent->position);
	}

	RemoveEntity(main_ent);
}

void StartPlayer(Entity *e)
{
	e->model = GetModel(&main_engine->modelChain, "WorldData/Models/ship.mdl");
	e->c_model = GetCModel(&main_engine->cmodelChain, "WorldData/Models/c_ship.mdl");
	e->texture = GetTexture(&main_engine->skinChain, "WorldData/Skins/ship.bmp");

	e->Touch = Player_Touch;
	e->Think = Player_Think;
	e->position = Vector(0,0,0);

	e->bbox = e->model->bbox;
	strcpy(e->type, "player");
}

/* ROCK FUNCTIONS */


#include "const.h"
void StartRock(Entity *e)
{
	e->model = GetModel(&main_engine->modelChain, "WorldData/Models/rock.mdl");
	e->c_model = GetCModel(&main_engine->cmodelChain, "WorldData/Models/c_rock.mdl");
	e->texture = GetTexture(&main_engine->skinChain, "WorldData/Skins/rock.bmp");

	e->color.l = 0xFF3F3F3F;

	e->rotation.x = MCRand();
	e->rotation.y = MCRand();
	e->rotation.z = MCRand();
	e->rotation.w = 0;
	e->rotation.Normalize();
	e->rotation.w = MCRand();
	e->rotation.AngleAxisToQuat();
	e->velocity = Vector(MCRand(), MCRand(), MCRand());

	e->scale = 20.f;

	e->bbox = e->model->bbox;

	strcpy(e->type, "rock_l");

	e->position = Vector(MCRand(), MCRand(), MCRand());
	e->position.scale(WorldSize);
}

/* SYSTEM FUNCTIONS */

#include "collision.h"


void UpdateEntities(void)
{
	int a,v;

	//clear 'touched' flag - debug display only
	for (a = 0; a < MAX_ENTS; a++)
		entityarray[a].touched = false;

	//update state space / think
	for (a = 0; a < MAX_ENTS; a++)
	{
		if (!entityarray[a].active) continue;
		main_ent = &entityarray[a];

		//execute the think function
		if (main_ent->Think) main_ent->Think();
	}

	if (e_UpdateCollision) Entity_UpdateCollision();

	for (a = 0; a < MAX_ENTS; a++)
	{
		if (!entityarray[a].active) continue;

		main_ent = &entityarray[a];

		//update axii
		for (v = 0; v < 3; v++) main_ent->angle.Rotate(&base_axis[v], &main_ent->axis[v]);
		//modulate position - check the bounds to the world
		
		for (int v = 0; v < 3; v++)
		{
			while (main_ent->position.v[v] > WorldSize) main_ent->position.v[v] -= 2 * WorldSize;
			while (main_ent->position.v[v] < -WorldSize) main_ent->position.v[v] += 2 * WorldSize;
		}
	}
}
