#include "geometry.h"
#include "const.h"

void Plane::CalculateEquation(double *output)
{
	output[0] = normal.x;
	output[1] = normal.y;
	output[2] = normal.z;
	output[3] = -normal.dot(&pos);
}

float Plane::Distance(Vector *v)
{
	Vector del;
	del.sub(v, &pos);
	return normal.dot(&del);
}

void Plane::ClosestPoint(Vector *in, Vector *out)
{
	Vector off;
	off.scale(&normal, Distance(in));
	out->sub(in, &off);
}

float Plane::LineIntersect(Line *lin, Vector *vout)
{
	float fraction = LineIntersectFraction(lin);
	vout->scale(&lin->dir, fraction);
	vout->add(&lin->pos);
	return fraction;
}

float Plane::LineIntersectFraction(Line *lin)
{
	float src_dist, dest_dist;

	src_dist = Distance(&lin->pos);

	Vector dest;
	dest.add(&lin->pos, &lin->dir);
	dest_dist = Distance(&dest);

	return src_dist / (src_dist - dest_dist);
}



void Line::PlaneIntersect(Plane *p1, Plane *p2)
{
	//the line dir is the cross of the two normals
	dir.cross(&p1->normal, &p2->normal);

	//keep 'em unit vectors
	dir.normalize();

	//the line pos is any point where the planes intersect
	//to find where a point where the two planes intersect
	//create a line from plane p1's origin to plane p2's origin
	//project the line onto plane p1
	//and trace it to intersect p2

	Line line_del;

	line_del.pos = p1->pos;

	line_del.dir.cross(&dir, &p1->normal);
	line_del.dir.normalize();

	p2->LineIntersect(&line_del, &pos);
}



void Box::Stretch(Vector *s_min, Vector *s_max)
{
	for (int n = 0; n < 3; n++)
	{
		if (s_min->v[n] < min.v[n]) min.v[n] = s_min->v[n];
		if (s_max->v[n] > max.v[n]) max.v[n] = s_max->v[n];
	}
}

float Box::Length(void)
{
	Vector del;
	del.sub(&max, &min);
	return del.length();
}

Vector Box::Center(void)
{
	Vector sum;
	sum.add(&min, &max);
	sum.scale(0.5f);
	return sum;
}

bool Box::Contains(Vector *v)
{
	return (v->x >= min.x && v->x <= max.x &&
			v->y >= min.y && v->y <= max.y &&
			v->z >= min.z && v->z <= max.z);
}

bool Box::Contains(Box *b)
{
	return (b->min.x >= min.x && b->max.x <= max.x &&
			b->min.y >= min.y && b->max.y <= max.y &&
			b->min.z >= min.z && b->max.z <= max.z);
}

bool Box::CheckLineIntersect(Vector *src, Vector *delta)
{
	Vector *vtx[2] = {&min, &max};
	Vector intersect;
	int v, d, di, dii;

	for (v = 0; v < 2; v++)
	{
		for (d = 0; d < 3; d++)
		{
			di = (d + 1) % 3;
			dii = (di + 1) % 3;

			if (delta->v[d])
			{
				float frac = (vtx[v]->v[d] - src->v[d]) / delta->v[d];
				if (frac < 0 || frac > 1) continue;
				intersect.scale(delta, frac);
				intersect.add(src);
			}
			else
			{
				intersect = *src;
			}

			if (intersect.v[di] >= min.v[di] && intersect.v[di] <= max.v[di] &&
				intersect.v[dii] >= min.v[dii] && intersect.v[dii] <= max.v[dii])
				return true;
		}
	}
	return false;
}

bool Box::Intersects(Box *b)
{
	return ((min.x < b->max.x) &&
			(min.y < b->max.y) &&
			(min.z < b->max.z) &&
			(max.x > b->min.x) &&
			(max.y > b->min.y) &&
			(max.z > b->min.z));
}

float Box::CheckBoxIntersect(Box *other, Vector del, Vector *res)
{
	/*
	Vector del = main_ent->velocity.v[v];
	for v = all three axii
		if (del.v[v] > 0)
			d = distance from main_ent->maxbbox.max.v[v] to e->maxbbox.min.v[v]
		else if (del.v[v] < 0)
			d = distance from main_ent->maxbbox.min.v[v] to e->maxbbox.max.v[v]
		f = d / del.v[v]		//fraction of movement to intersection
		if (f < 0 || f > 1) continue	//make sure we're not travelling beyond velocity vector
		increment box position by scaled velocity
		check for box intersection crossing on other two dimensions vi, vii
		if intersection found
			set intersection vector to the colliding vertex on your box [vi,vii]
			set movement fraction to 'f'
			scale del by 'f'
	*/

	int v,dim, di, dii;
	float dist, frac;
	Box newcheck;
	int src_id, dest_id;
	Vector del_scale, intersection;
	float *f_di, *f_dii;

	float res_frac = 1.f;

	for (dim = 0; dim < 3; dim++)
	{
		if (del.v[v] == 0) continue;
		if (del.v[v] > 0)
		{
			src_id = 1;
			dest_id = 0;
		}
		if (del.v[v] < 0) 
		{
			src_id = 0;
			dest_id = 1;
		}
		dist = other->vec()[dest_id].v[v] - this->vec()[src_id].v[v];
		frac = dist / del.v[v];
		if (frac < 0 || frac > 1) continue;		//no intersection found along del line segment

		del_scale.scale(&del, frac);
		newcheck.min.add(&this->min, &del_scale);
		newcheck.max.add(&this->max, &del_scale);

		di = (dim+1)%3;
		dii = (dim+2)%3;

		//if intersection is found
		if ((newcheck.min.v[di] < other->max.v[di]) &&
			(newcheck.max.v[di] > other->min.v[di]) &&
			(newcheck.min.v[dii] < other->max.v[dii]) &&
			(newcheck.max.v[dii] > other->min.v[dii]))
		{
			//grab the box corner of src
			if (newcheck.min.v[di] > other->min.v[di] && newcheck.min.v[di] < other->max.v[di])
				f_di = &newcheck.min.v[di];
			else if (newcheck.max.v[di] > other->min.v[di] && newcheck.max.v[di] < other->max.v[di])
				f_di = &newcheck.max.v[di];
			else
				f_di = &other->min.v[di];	//any vertex of other - doesnt matter
			
			if (newcheck.min.v[dii] > other->min.v[dii] && newcheck.min.v[dii] < other->max.v[dii])
				f_dii = &newcheck.min.v[dii];
			else if (newcheck.max.v[dii] > other->min.v[dii] && newcheck.max.v[dii] < other->max.v[dii])
				f_dii = &newcheck.max.v[dii];
			else
				f_dii = &other->min.v[dii];

			//corner min.v[di], min.v[dii]
			intersection.v[dim] = newcheck.vec()[src_id].v[dim];	//apply this to 
			intersection.v[di] = *f_di;
			intersection.v[dii] = *f_dii;

			del = del_scale;
			res_frac *= frac;
		}
	}

	if (res_frac < 1.f) *res = intersection;

	return res_frac;
}