#include <windows.h>
#include <gl/gl.h>

#include "model.h"

Model::Model()
{
	triCount = 0;
	vtxCount = 0;

	tri = NULL;
	vtx = NULL;
	texCoord = NULL;
	normal = NULL;
}

#include <stdio.h>
#include "error.h"
#include "const.h"
bool Model::Load(char *filename)
{
	//CUSTOM MDL FILE FORMAT

	int t,v;

	strcpy(title, filename);

	char sig[4];

	FILE *File = fopen(filename, "rb");
	if (File == NULL)
	{
		Error("%s: unable to open model file", filename);
		return false;
	}

	fread(sig, 1, 4, File);
	if (memcmp(sig, "MMDL", 4))
	{
		Error("%s: bad signature on file", filename);
		return false;
	}
	fread(&vtxCount, sizeof(int), 1, File);
	fread(&triCount, sizeof(int), 1, File);

	if (vtxCount < 0)
	{
		Error("%s: model has bad vertex count: %d", filename, vtxCount);
		fclose(File);
		return false;
	}
	if (triCount < 0)
	{
		Error("%s: model has bad triangle count: %d", filename, triCount);
		fclose(File);
		return false;
	}

	if ((vtx = (Vector *)malloc(sizeof(Vector) * vtxCount))==NULL)
	{
		Error("%s: unable to allocate vertex array", filename);
		fclose(File);
		return false;
	}
	if ((texCoord = (texcoord_t *)malloc(sizeof(texcoord_t) * vtxCount))==NULL)
	{
		Error("%s: unable to allocate texture coordinate array", filename);
		free(vtx);
		fclose(File);
		return false;
	}
	if ((normal = (Vector *)malloc(sizeof(Vector) * vtxCount))==NULL)
	{
		Error("%s: unable to allocate normal array", filename);
		free(texCoord);
		free(vtx);
		fclose(File);
		return false;
	}
	if ((tri = (int *)malloc(sizeof(int) * 3 * triCount))==NULL)
	{
		Error("%s: unable to allocate triangle array", filename);
		free(normal);
		free(texCoord);
		free(vtx);
		fclose(File);
		return false;
	}

	if (fread(vtx, sizeof(Vector), vtxCount, File) != (size_t)(vtxCount))
	{
		Error("%s: unable to read vertex array", filename);
		fclose(File);
		return false;
	}
	for (v = 0; v < vtxCount; v++)
	{
		float tc[2];
		if (fread(&tc, sizeof(float), 2, File) != 2)
		{
			Error("%s: unable to read texture coordinate array", filename);
			fclose(File);
			return false;
		}
		texCoord[v].st[0] = (short)(tc[0]*256);	//fixed-point-ize it
		texCoord[v].st[1] = (short)(tc[1]*256);
	}
	if (fread(tri, sizeof(int), triCount*3, File) != (size_t)(triCount*3))
	{
		Error("%s: unable to read triangle array", filename);
		fclose(File);
		return false;
	}

	fclose(File);

	//calc normals
	for (v = 0; v < vtxCount; v++) normal[v] = Vector(0,0,0);

	//flip tris
	for (t = 0; t < triCount*3; t+=3)
	{
		int tmp = tri[t];
		tri[t] = tri[t+2];
		tri[t+2] = tmp;

		Vector n;
		n.CalcNormal(&vtx[tri[t]], &vtx[tri[t+1]], &vtx[tri[t+2]]);

		for (int i = 0; i < 3; i++) normal[tri[t+i]] += n;
	}

	bbox.min = Vector(fInf,fInf,fInf);
	bbox.max = Vector(-fInf,-fInf,-fInf);
	radius = -1;

	for (v = 0; v < vtxCount; v++)
	{
		//normalize normals
		normal[v].normalize();

		//and stretch bbox
		bbox.Stretch(&vtx[v]);

		float l = vtx[v].length();
		if (l > radius) radius = l;
	}

	return true;
}

void Model::Unload(void)
{
	triCount = 0;
	vtxCount = 0;

	if (tri) free(tri);
	tri = NULL;

	if (vtx) free(vtx);
	vtx = NULL;

	if (texCoord) free(texCoord);
	texCoord = NULL;

	if (normal) free(normal);
	normal = NULL;
}

void DisplayModel(Model *m)
{
	if (!m) return;

	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);

	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);

	glVertexPointer(3, GL_FLOAT, 0, m->vtx);
	glTexCoordPointer(2, GL_SHORT, 0, m->texCoord);
	glNormalPointer(GL_FLOAT, 0, m->normal);

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);

	glDrawElements(GL_TRIANGLES, 3*m->triCount, GL_UNSIGNED_INT, m->tri);

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);

	//reset changes
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
}

Model *GetModel(Chain *p, char *title)
{
	Model *m = (Model *)p->ChainHead();
	for (; m; m = (Model *)m->ChainNext())
		if (!_stricmp(title, m->title)) break;

	return m;
}
