#include "particle.h"
#include "texture.h"
#include "quat.h"

#include <windows.h>
#include <gl/gl.h>

unsigned char particle_texture[8][8] =
{

{0,0,50,55,50,0,0,0},
{0,60,84,98,84,60,0,0},
{50,84,141,182,141,84,50,0},
{55,98,182,255,182,98,55,0},
{50,84,141,182,141,84,50,0},
{0,60,84,98,84,60,0,0},
{0,0,50,55,50,0,0,0},
{0,0,0,0,0,0,0,0}

};

bool ParticleSystem::Init(int max)
{
	//set the count
	particleCount = max;
	
	//allocate the particle array
	particle = (Particle *)malloc(sizeof(Particle) * particleCount);
	if (!particle) return false;

	for (int a = 0; a < particleCount; a++)
	{
		memset(&particle[a], 0, sizeof(Particle));	//zero it
		particle[a].Link(&deadChain);				//link it
	}

	//init the texture chain
	Texture *t = new Texture();
	if (!t->Create("WorldData/smoke.tga")) return false;
	t->Link(&textureChain);

	return true;
}

void ParticleSystem::Kill(void)
{
	//unload
	liveChain.UnlinkAll();
	deadChain.UnlinkAll();
	textureChain.DeleteAll();
	if (particle) free(particle);
	//and reset
	Reset();
}

void ParticleSystem::Reset(void)
{
	particle = NULL;
	particleCount = 0;

	last_blendfunc.st[0] = GL_SRC_ALPHA;
	last_blendfunc.st[1] = GL_ONE;
}

void ParticleSystem::SetBlend(int src, int dest)
{
	last_blendfunc.st[0] = src;
	last_blendfunc.st[1] = dest;
}

void ParticleSystem::Update(void)
{
	Particle *next;

	for (Particle *p = (Particle *)liveChain.ChainHead(); p; p = next)
	{
		next = (Particle *)p->ChainNext();

		p->life--;
		if (p->life <= 0)
		{
			p->Unlink();
			p->Link(&deadChain);
			continue;
		}
		p->position.add(&p->velocity);

	}
}

void ParticleSystem::LinkParticle(Particle *p, Particle *node)
{
	int ch = p->dist > node->dist;

	if (node->child[ch])
		LinkParticle(p, node->child[ch]);
	else
	{
		node->child[ch] = p;
		p->child[0] = p->child[1] = NULL;
	}
}

void ParticleSystem::DisplayParticle(Particle *p)
{
	if (p->child[0]) DisplayParticle(p->child[0]);

	float size = p->size * 0.5f;
	color_t		color;

	color.l = p->color.l;
	color.rgba[3] = (unsigned char)( (float)p->color.rgba[3] * (float)p->life / (float)p->max_life);
	glColor4ubv(color.rgba);

	glBlendFunc(p->blendfunc.st[0], p->blendfunc.st[1]);

	glBegin(GL_TRIANGLE_STRIP);
	glTexCoord2f(0,0);
	glVertex3f(	p->position.x - size * (view_left.x + view_up.x),
				p->position.y - size * (view_left.y + view_up.y),
				p->position.z - size * (view_left.z + view_up.z));
	glTexCoord2f(1,0);
	glVertex3f(	p->position.x - size * (view_left.x - view_up.x),
				p->position.y - size * (view_left.y - view_up.y),
				p->position.z - size * (view_left.z - view_up.z));
	glTexCoord2f(0,1);
	glVertex3f(	p->position.x + size * (view_left.x - view_up.x),
				p->position.y + size * (view_left.y - view_up.y),
				p->position.z + size * (view_left.z - view_up.z));
	glTexCoord2f(1,1);
	glVertex3f(	p->position.x + size * (view_left.x + view_up.x),
				p->position.y + size * (view_left.y + view_up.y),
				p->position.z + size * (view_left.z + view_up.z));
	glEnd();

	if (p->child[1]) DisplayParticle(p->child[1]);
}

void ParticleSystem::Display(M_Camera *camera)
{
	view_left = camera->axis[1];
	view_up = camera->axis[2];

	Particle *p, *root;
	Vector del;

	root = (Particle *)liveChain.ChainHead();
	if (!root) return;

	//calc depth
	for (p = root; p; p = (Particle *)p->ChainNext())
	{
		del.sub(&p->position, &camera->position);
		p->dist = del.normal();
	}

	//build binary tree
	p = root;
	p->child[0] = p->child[1] = NULL;
	for (p = (Particle *)p->ChainNext(); p; p = (Particle *)p->ChainNext())
		LinkParticle(p, (Particle *)liveChain.ChainHead());

	glEnable(GL_BLEND);
	glDepthMask(GL_FALSE);

	SetTexture((Texture *)textureChain.ChainHead());

	DisplayParticle(root);

	glDepthMask(GL_TRUE);
	glDisable(GL_BLEND);
}

Particle *ParticleSystem::NewParticle(void)
{
	Particle *p = (Particle *)deadChain.ChainTail();
	if (p)
	{
		p->Unlink();
		p->Link(&liveChain);
	}
	return p;
}

void ParticleSystem::CreateParticle(Vector *pos, Vector *vel, float size, int color, int life)
{
	Particle *p;
	p = NewParticle();
	if (!p) return;

	p->position = *pos;
	p->velocity = *vel;
	p->size = size;
	p->color.l = color;
	p->life = p->max_life = life;
	p->blendfunc.l = last_blendfunc.l;
}