#include <windows.h>

#include "error.h"
#include "oglh.h"

#include "display.h"
#include "keyboard.h"

HINSTANCE My_RegisterClass(WNDPROC wndProc)
{
    WNDCLASS    wc;
	static HINSTANCE hInstance = 0;

    /* only register the window class once - use hInstance as a flag. */
    if (!hInstance) {
		hInstance = GetModuleHandle(NULL);

		wc.style         = CS_OWNDC;
		wc.lpfnWndProc   = wndProc;
		wc.cbClsExtra    = 0;
		wc.cbWndExtra    = 0;
		wc.hInstance     = hInstance;
		wc.hIcon         = LoadIcon(NULL, IDI_WINLOGO);
		wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
		wc.hbrBackground = NULL;
		wc.lpszMenuName  = NULL;
		wc.lpszClassName = "OpenGL";

		if (!RegisterClass(&wc)) {
		    Error("My_RegisterClass(): RegisterClass() failed:  ");
		    return 0;
		}
    }

	return hInstance;
}

HWND My_CreateWindow(int x, int y, int width, int height, char *title, WNDPROC wndProc)
{
    HWND hWnd;
	HINSTANCE hInst;

	if ( (hInst = My_RegisterClass(wndProc)) == 0)
	{
		Error("My_CreateWindow(): My_RegisterClass() failed");
		return NULL;
	}

	DWORD		dwExStyle;				// Window Extended Style
	DWORD		dwStyle;				// Window Style

	dwExStyle=WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;						// Window Extended Style
	dwStyle=WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;	// Windows Style

    hWnd = CreateWindowEx(dwExStyle, "OpenGL", title,
			dwStyle, x, y, width, height, NULL, NULL, hInst, NULL);

    if (hWnd == NULL)
	{
		Error("My_CreateWindow():  hWnd failed to be created");
		return NULL;
    }

	return hWnd;
}

LONG WINAPI DisplayWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	static OpenGLHandler *oglH = NULL;
	static M_KeyboardHandler *kbdH = NULL;
	static PAINTSTRUCT ps;
	static void (*menufunc)(int menuID, void *param) = NULL;
	static void *menufunc_param = NULL;

	switch(uMsg) {

	case WM_CREATE:
		return 0;

	case MY_WM_POST_OGLH:
		oglH = (OpenGLHandler *)lParam;
		return 0;
	case MY_WM_POST_KBDH:
		kbdH = (M_KeyboardHandler *)lParam;
		return 0;
	case MY_WM_POST_MENUFUNC:
		menufunc = (void (*)(int,void *))wParam;
		menufunc_param = (void *)lParam;
		return 0;

	case WM_PAINT:
		BeginPaint(hWnd, &ps);
		EndPaint(hWnd, &ps);
		return 0;

	case WM_SIZE:
		//post a resize message
		if (oglH)
		{
			oglH->Resize();
			oglH->Clear();
		}
		return 0;

	case WM_KEYDOWN:
		if (kbdH) kbdH->KeyDown(wParam);
		return 0;
	case WM_KEYUP:
		if (kbdH) kbdH->KeyUp(wParam);
		return 0;

	case WM_COMMAND:	//interpret menu commands

		//if its a menu being clicked and there is a function to trap it
		if ((HIWORD(wParam) == 0) && menufunc)
			//then execute the function and pass the menu ID through
			menufunc(LOWORD(wParam), menufunc_param);

		return 0;


	case WM_CLOSE:
		//post a quit message for the main program
		PostQuitMessage(0);	//mayhap we should use a different method of notifying a global exit?
		//and go
		return 0;
	}

	return DefWindowProc(hWnd, uMsg, wParam, lParam); 
} 

M_Display::M_Display()
{
	//NULL the pointers
	displayWnd = NULL;
}

bool M_Display::Init(void)
{
	//create the display window and pass the memory lcoation of the opengl handler pointer
	//to the window.  we are passing the pointer to the pointer so that DisplayWindowProc
	//will have the abillity to modify the variable's data.  this ability is needed
	//if the Initialization of the opengl handler fails, and the pointer needs to be set
	//to NULL
	displayWnd = My_CreateWindow(300,300,640,480, "Display Window", DisplayWindowProc);
	if (displayWnd == NULL)
	{
		Error("InitDisplay() displayWnd failed to be created");
		return false;
	}

	if (oglh.Init(displayWnd, true) == false)	//if the opengl display obj failed to Init
	{
		//destroy opengl handler
		oglh.Kill();
		return false;	//WM_CREATE failed
	}

	//good to go - show the window and leave
	ShowWindow(displayWnd, SW_SHOW);

	//send the OpenGLHandler pointer to the window - for catching WM_RESIZE messages
	SendMessage(displayWnd, MY_WM_POST_OGLH, 0, (LPARAM)&oglh);

	return true;
}

//send the OpenGLHandler pointer to the window - for catching WM_RESIZE messages
void M_Display::SetKeyboardHandler(M_KeyboardHandler *kbdH)
{
	SendMessage(displayWnd, MY_WM_POST_KBDH, 0, (LPARAM)kbdH);
}

void M_Display::Update(void)
{
	//update the opengl handler (if we have an active one)
	oglh.Update();
}

void M_Display::Clear(void)
{
	oglh.Clear();
}


M_Display::~M_Display() { Kill(); }

void M_Display::Kill(void)
{
	//if the OpenGL Handler exists, then kill it!
	oglh.Kill();

	//same to you, display window!
	if (displayWnd)
	{
		DestroyWindow(displayWnd);
		displayWnd = NULL;
	}
}

int M_Display::GetWidth(void) {	return oglh.Width();}
int M_Display::GetHeight(void){ return oglh.Height();}
